// SignSymbol.java
// Version 1.0 by DTAI
// Version 1.1 by Larry S. Peranich 12-17-98
//      Change to use new symbol order given in newest edition of symbol file
//      new symbol file has symbols in Sign-Symbol-Sequence, which is 
//      "alphabetical order" for sign symbols
// Version 1.2 by Larry S. Peranich 12-27-98
//      Changed to put all old-symbol-code to new-symbol-code translations in
//      the SignSymbol(int code, int opt, byte y, byte x, boolean version4_0)
//      function.  This function is used by both SWFile.java (reads in old, 
//      version 4.3 documents) and the dictionary code (reads in old, version 4.3
//      dictionary files), so now the dictionaries work with the new symbol order.
//
//      In addition, implement phase 2 of the old-symbol-code to new-symbol-code 
//      translation.  Phase 1 only affected symbol codes (as mapped in mapOldToNew),
//      but phase 2 maps special cases old(symbol code, variation, rotation, fill) to
//      new(symbol code, variation, rotation, fill).
// Version 1.3 by Larry S. Peranich 1-11-99
//      Implement phase 3 of the old-symbol-code to new-symbol-code translation.  
//      Phase 3 uses a new hash table, that maps symcode,fill,ahead values, but 
//      leaves rotations unchanged (unlike phase 2).  Some of the translations
//      originally in phase 2 were move to the phase 3 hash table because they
//      were the same across all rotations.

package signwriting;

import dtai.gwt.GadgetGraphics;
import java.awt.Dimension;
import java.awt.Point;
import java.util.Hashtable;

public class SignSymbol extends SignComponent {
    
    static final long serialVersionUID = 4927791797228843442L;
    
    private static Hashtable hash1;
    
    private static Hashtable phase2Map;
    
    private static Hashtable phase3Map;

    // array to map the old codes to the new codes (phase 1 of mapping)
    // newCode = mapOldToNew[oldCode];
    // LSP 12-17-98
    private static final int mapOldToNew[] = {
          1,   2,   5,   6,   7,   8,   9,  10,  11,  12,
         15,  16,  17,  18,  19,  20,  21,  22,  23,  25,
         26,  27,  39,  28,  29,  31,  37,  32,  51,  35,
         52,  43,  33,  44,  45,  50,  49,  48,  46,  53,
         55,  56,  57,  58,  59,  61,  62,  63,  64,  65,
         66,  67,  68,  69,  70,  72,  73,  74,  75,  76,
         82,  79,  83,  81,  85,  80,  84,  87,  88,  89,
         93,  91,  94,  92,  98,  96,  97, 100, 101, 102,
        103, 104, 106,  99,   3,  90,  95, 108, 109, 110,
        111, 112, 113, 114, 115, 116, 117, 118, 119, 120,
        121, 122, 123, 124, 125, 126, 127, 128, 129, 130,
        131, 132, 133, 134, 135, 136, 137, 138, 139, 140,
        141, 142, 143, 144, 145, 146, 147, 148, 149, 151,
        152, 153, 154, 155, 150, 156, 157, 158, 159, 160,
        161, 162, 163, 164, 165, 166, 167, 168, 169, 170,
        171, 172, 173, 174, 175, 176, 177, 178, 179, 180,
        181, 182, 183, 184, 185, 186, 187, 188, 189, 190,
        193, 194, 195, 196, 197, 198, 199, 200, 201, 202,
        203, 204, 205, 191,  36, 206, 207,  30, 208, 209,
         71, 210, 211, 212, 213, 214,  38, 215,  34, 216,
        217,  42,  41,  40,  13, 218,  14,  60, 219, 220,
         47, 221, 222,  78,  77, 223, 224, 225,  86,  24,
        226, 227, 228, 229, 230, 231, 232, 233, 234, 235,
        236, 237, 238, 107, 239, 240, 241, 105, 242,  54,
        243, 244, 245, 246, 247, 192, 248, 249, 250, 251
    };
    
    // An array that is used to fill in the phase2Map hashtable
    // Each line consists of 
    // { old_code, old_fill, old_rot, old_var, new_code, new_fill,
    //      new_rot, new_var}
    // NOTE:  the values of fill, rot (rotation) and var (variation or ahead)
    //      in this table start counting at one, while the actual values start
    //      counting at zero.  Subtract one from all fill, rot, and var values
    //      when constructing the hashtable.  These values are the ones Valerie
    //      sees when using her symbol table editor.
    private static final int phase2MapInit[][] = {
        {108, 1, 2, 1, 108, 1, 1, 2},
        {109, 1, 2, 1, 109, 1, 1, 2},
        {110, 1, 2, 1, 110, 1, 1, 2},
        {111, 1, 2, 1, 111, 1, 1, 2},
        {112, 1, 2, 1, 112, 1, 1, 2},
        {113, 1, 2, 1, 113, 1, 1, 2},
        {114, 1, 2, 1, 114, 1, 1, 2},
        {115, 1, 2, 1, 115, 1, 1, 2},
        {152, 1, 2, 1, 152, 1, 1, 2},
        {152, 2, 2, 1, 152, 2, 1, 2},
        {152, 3, 2, 1, 152, 3, 1, 2},
        {152, 1, 3, 1, 152, 1, 1, 3},
        {152, 2, 3, 1, 152, 2, 1, 3},
        {152, 3, 3, 1, 152, 3, 1, 3},
        {153, 1, 2, 1, 153, 1, 1, 2},
        {153, 2, 2, 1, 153, 2, 1, 2},
        {153, 3, 2, 1, 153, 3, 1, 2},
        {153, 1, 3, 1, 153, 1, 1, 3},
        {153, 2, 3, 1, 153, 2, 1, 3},
        {153, 3, 3, 1, 153, 3, 1, 3},
        {154, 1, 2, 1, 154, 1, 1, 2},
        {154, 2, 2, 1, 154, 2, 1, 2},
        {154, 3, 2, 1, 154, 3, 1, 2},
        {154, 1, 3, 1, 154, 1, 1, 3},
        {154, 2, 3, 1, 154, 2, 1, 3},
        {154, 3, 3, 1, 154, 3, 1, 3},
        {155, 1, 2, 1, 155, 1, 1, 2},
        {155, 2, 2, 1, 155, 2, 1, 2},
        {155, 3, 2, 1, 155, 3, 1, 2},
        {155, 1, 3, 1, 155, 1, 1, 3},
        {155, 2, 3, 1, 155, 2, 1, 3},
        {155, 3, 3, 1, 155, 3, 1, 3},
        {156, 1, 1, 1, 156, 1, 1, 2},
        {156, 2, 1, 1, 156, 2, 1, 2},
        {156, 3, 1, 1, 156, 1, 1, 1},
        {156, 4, 1, 1, 156, 2, 1, 1},
        {157, 1, 2, 1, 157, 1, 1, 2},
        {157, 2, 2, 1, 157, 2, 1, 2},
        {157, 3, 2, 1, 157, 3, 1, 2},
        {157, 1, 3, 1, 157, 1, 1, 3},
        {157, 2, 3, 1, 157, 2, 1, 3},
        {157, 3, 3, 1, 157, 3, 1, 3},
        {158, 1, 2, 1, 158, 1, 1, 2},
        {158, 2, 2, 1, 158, 2, 1, 2},
        {158, 3, 2, 1, 158, 3, 1, 2},
        {158, 1, 3, 1, 158, 1, 1, 3},
        {158, 2, 3, 1, 158, 2, 1, 3},
        {158, 3, 3, 1, 158, 3, 1, 3},
        {159, 1, 2, 1, 159, 1, 1, 2},
        {159, 2, 2, 1, 159, 2, 1, 2},
        {159, 3, 2, 1, 159, 3, 1, 2},
        {159, 1, 3, 1, 159, 1, 1, 3},
        {159, 2, 3, 1, 159, 2, 1, 3},
        {159, 3, 3, 1, 159, 3, 1, 3},
        {160, 1, 2, 1, 160, 1, 1, 2},
        {160, 2, 2, 1, 160, 2, 1, 2},
        {160, 3, 2, 1, 160, 3, 1, 2},
        {160, 1, 3, 1, 160, 1, 1, 3},
        {160, 2, 3, 1, 160, 2, 1, 3},
        {160, 3, 3, 1, 160, 3, 1, 3},
        {166, 1, 2, 1, 166, 1, 1, 2},
        {166, 2, 2, 1, 166, 2, 1, 2},
        {166, 3, 2, 1, 166, 3, 1, 2},
        {166, 1, 3, 1, 166, 1, 1, 3},
        {166, 2, 3, 1, 166, 2, 1, 3},
        {166, 3, 3, 1, 166, 3, 1, 3},
        {170, 1, 2, 1, 170, 1, 1, 2},
        {170, 2, 2, 1, 170, 2, 1, 2},
        {170, 3, 2, 1, 170, 3, 1, 2},
        {170, 1, 3, 1, 170, 1, 1, 3},
        {170, 2, 3, 1, 170, 2, 1, 3},
        {170, 3, 3, 1, 170, 3, 1, 3},
        {175, 1, 2, 1, 175, 1, 1, 2},
        {175, 2, 2, 1, 175, 2, 1, 2},
        {175, 3, 2, 1, 175, 3, 1, 2},
        {175, 1, 3, 1, 175, 1, 1, 3},
        {175, 2, 3, 1, 175, 2, 1, 3},
        {175, 3, 3, 1, 175, 3, 1, 3},
        {179, 1, 2, 1, 179, 1, 1, 2},
        {179, 2, 2, 1, 179, 2, 1, 2},
        {179, 3, 2, 1, 179, 3, 1, 2},
        {179, 1, 3, 1, 179, 1, 1, 3},
        {179, 2, 3, 1, 179, 2, 1, 3},
        {179, 3, 3, 1, 179, 3, 1, 3},
        {181, 1, 2, 1, 181, 1, 1, 2},
        {181, 2, 2, 1, 181, 2, 1, 2},
        {181, 3, 2, 1, 181, 3, 1, 2},
        {181, 1, 3, 1, 181, 1, 1, 3},
        {181, 2, 3, 1, 181, 2, 1, 3},
        {181, 3, 3, 1, 181, 3, 1, 3},
        {182, 1, 2, 1, 182, 1, 1, 2},
        {182, 2, 2, 1, 182, 2, 1, 2},
        {182, 3, 2, 1, 182, 3, 1, 2},
        {182, 1, 3, 1, 182, 1, 1, 3},
        {182, 2, 3, 1, 182, 2, 1, 3},
        {182, 3, 3, 1, 182, 3, 1, 3},
        {183, 1, 2, 1, 183, 1, 1, 2},
        {183, 2, 2, 1, 183, 2, 1, 2},
        {183, 3, 2, 1, 183, 3, 1, 2},
        {183, 1, 3, 1, 183, 1, 1, 3},
        {183, 2, 3, 1, 183, 2, 1, 3},
        {183, 3, 3, 1, 183, 3, 1, 3},
        {184, 1, 2, 1, 184, 1, 1, 2},
        {184, 2, 2, 1, 184, 2, 1, 2},
        {184, 3, 2, 1, 184, 3, 1, 2},
        {184, 1, 3, 1, 184, 1, 1, 3},
        {184, 2, 3, 1, 184, 2, 1, 3},
        {184, 3, 3, 1, 184, 3, 1, 3},
        {186, 1, 2, 1, 186, 1, 1, 2},
        {186, 2, 2, 1, 186, 2, 1, 2},
        {186, 3, 2, 1, 186, 3, 1, 2},
        {186, 1, 3, 1, 186, 1, 1, 3},
        {186, 2, 3, 1, 186, 2, 1, 3},
        {186, 3, 3, 1, 186, 3, 1, 3},
        {187, 1, 2, 1, 187, 1, 1, 2},
        {187, 2, 2, 1, 187, 2, 1, 2},
        {187, 3, 2, 1, 187, 3, 1, 2},
        {187, 1, 3, 1, 187, 1, 1, 3},
        {187, 2, 3, 1, 187, 2, 1, 3},
        {187, 3, 3, 1, 187, 3, 1, 3},
        {188, 1, 2, 1, 188, 1, 1, 2},
        {188, 2, 2, 1, 188, 2, 1, 2},
        {188, 3, 2, 1, 188, 3, 1, 2},
        {188, 1, 3, 1, 188, 1, 1, 3},
        {188, 2, 3, 1, 188, 2, 1, 3},
        {188, 3, 3, 1, 188, 3, 1, 3},
        {189, 1, 2, 1, 189, 1, 1, 2},
        {189, 2, 2, 1, 189, 2, 1, 2},
        {189, 3, 2, 1, 189, 3, 1, 2},
        {189, 1, 3, 1, 189, 1, 1, 3},
        {189, 2, 3, 1, 189, 2, 1, 3},
        {189, 3, 3, 1, 189, 3, 1, 3},
        {199, 1, 2, 1, 199, 1, 1, 2},
        {199, 1, 4, 1, 199, 1, 3, 2},
        {199, 2, 2, 1, 199, 2, 1, 2},
        {199, 2, 4, 1, 199, 2, 3, 2},
        {199, 3, 2, 1, 199, 3, 1, 2},
        {199, 3, 4, 1, 199, 3, 3, 2},
        {200, 1, 2, 1, 200, 1, 1, 2},
        {200, 2, 2, 1, 200, 2, 1, 2},
        {200, 3, 2, 1, 200, 3, 1, 2},
        {200, 1, 3, 1, 200, 1, 1, 3},
        {200, 2, 3, 1, 200, 2, 1, 3},
        {200, 3, 3, 1, 200, 3, 1, 3},
        {200, 1, 4, 1, 200, 1, 1, 4},
        {200, 2, 4, 1, 200, 2, 1, 4},
        {200, 3, 4, 1, 200, 3, 1, 4},
        {201, 1, 2, 1, 201, 1, 1, 2},
        {201, 2, 2, 1, 201, 2, 1, 2},
        {201, 3, 2, 1, 201, 3, 1, 2},
        {201, 1, 3, 1, 201, 1, 1, 3},
        {201, 2, 3, 1, 201, 2, 1, 3},
        {201, 3, 3, 1, 201, 3, 1, 3},
        {201, 1, 4, 1, 201, 1, 1, 4},
        {201, 2, 4, 1, 201, 2, 1, 4},
        {201, 3, 4, 1, 201, 3, 1, 4},
        {202, 1, 2, 1, 202, 1, 1, 2},
        {202, 2, 2, 1, 202, 2, 1, 2},
        {202, 3, 2, 1, 202, 3, 1, 2},
        {202, 1, 3, 1, 202, 1, 1, 3},
        {202, 2, 3, 1, 202, 2, 1, 3},
        {202, 3, 3, 1, 202, 3, 1, 3},
        {204, 2, 1, 1, 204, 1, 1, 2},
        {204, 3, 1, 1, 204, 1, 1, 3},
        {204, 4, 1, 1, 204, 1, 1, 4},
        {205, 2, 1, 1, 205, 1, 1, 2},
        {206, 1, 1, 1, 207, 1, 1, 1},
        {206, 2, 1, 1, 207, 2, 1, 1},
        {206, 1, 2, 1, 207, 1, 1, 2},
        {206, 2, 2, 1, 207, 2, 1, 2},
        {206, 1, 3, 1, 207, 1, 1, 3},
        {206, 2, 3, 1, 207, 2, 1, 3},
        {207, 1, 1, 1, 208, 1, 1, 1},
        {207, 2, 1, 1, 208, 2, 1, 1},
        {207, 1, 2, 1, 208, 1, 1, 2},
        {207, 2, 2, 1, 208, 2, 1, 2},
        {207, 1, 3, 1, 208, 1, 1, 3},
        {207, 2, 3, 1, 208, 2, 1, 3},
        {207, 1, 4, 1, 208, 1, 1, 4},
        {207, 2, 4, 1, 208, 2, 1, 4},
        {208, 1, 1, 1, 209, 1, 1, 1},
        {208, 2, 1, 1, 209, 2, 1, 1},
        {208, 1, 2, 1, 209, 1, 1, 2},
        {208, 2, 2, 1, 209, 2, 1, 2},
        {208, 1, 3, 1, 209, 1, 1, 3},
        {208, 2, 3, 1, 209, 2, 1, 3},
        {209, 1, 1, 1, 210, 1, 1, 1},
        {209, 2, 1, 1, 210, 2, 1, 1},
        {209, 1, 2, 1, 210, 1, 1, 2},
        {209, 2, 2, 1, 210, 2, 1, 2},
        {209, 1, 3, 1, 210, 1, 1, 3},
        {209, 2, 3, 1, 210, 2, 1, 3},
        {209, 1, 4, 1, 210, 1, 1, 4},
        {209, 2, 4, 1, 210, 2, 1, 4},
        {211, 1, 1, 1, 212, 1, 1, 1},
        {211, 2, 1, 1, 212, 2, 1, 1},
        {211, 1, 2, 1, 212, 1, 1, 2},
        {211, 2, 2, 1, 212, 2, 1, 2},
        {211, 1, 3, 1, 212, 1, 1, 3},
        {211, 2, 3, 1, 212, 2, 1, 3},
        {212, 1, 1, 1, 213, 1, 1, 1},
        {212, 2, 1, 1, 213, 2, 1, 1},
        {212, 1, 2, 1, 213, 1, 1, 2},
        {212, 2, 2, 1, 213, 2, 1, 2},
        {212, 1, 3, 1, 213, 1, 1, 3},
        {212, 2, 3, 1, 213, 2, 1, 3},
        {213, 1, 1, 1, 214, 1, 1, 1},
        {213, 1, 2, 1, 214, 1, 2, 1},
        {213, 1, 3, 1, 215, 1, 1, 1},
        {214, 1, 1, 1, 216, 1, 1, 1},
        {214, 1, 2, 1, 216, 1, 1, 2},
        {214, 2, 1, 1, 216, 1, 1, 3},
        {214, 2, 2, 1, 216, 1, 1, 4},
        {214, 3, 1, 1, 225, 1, 1, 2},
        {214, 3, 2, 1, 225, 1, 1, 3},
        {214, 4, 1, 1, 228, 1, 1, 5},
        {214, 4, 2, 1, 228, 1, 1, 6},
        {215, 1, 1, 1, 217, 1, 1, 1},
        {215, 1, 2, 1, 217, 1, 1, 2},
        {215, 2, 1, 1, 217, 1, 1, 3},
        {215, 2, 2, 1, 217, 1, 1, 4},
        {215, 3, 1, 1, 206, 1, 1, 1},
        {215, 3, 2, 1, 206, 1, 1, 2},
        {216, 1, 1, 1, 218, 1, 1, 1},
        {216, 1, 2, 1, 218, 1, 1, 2},
        {216, 2, 1, 1, 222, 1, 1, 1},
        {216, 2, 2, 1, 222, 1, 2, 1},
        {216, 3, 1, 1, 222, 1, 1, 2},
        {216, 3, 2, 1, 222, 1, 2, 2},
        {217, 1, 1, 1, 219, 1, 1, 1},
        {217, 1, 2, 1, 219, 1, 1, 2},
        {217, 1, 3, 1, 219, 1, 1, 3},
        {217, 2, 1, 1, 220, 1, 1, 1},
        {217, 2, 2, 1, 220, 1, 1, 2},
        {217, 2, 3, 1, 220, 1, 1, 3},
        {217, 3, 1, 1, 221, 1, 1, 1},
        {217, 3, 2, 1, 221, 1, 1, 2},
        {217, 3, 3, 1, 221, 1, 1, 3},
        {218, 1, 1, 1, 224, 1, 1, 1},
        {218, 1, 2, 1, 224, 1, 1, 2},
        {218, 1, 3, 1, 224, 1, 1, 3},
        {218, 2, 1, 1, 223, 1, 1, 1},
        {218, 2, 2, 1, 223, 1, 1, 2},
        {218, 2, 3, 1, 225, 1, 1, 1},
        {219, 1, 1, 1, 227, 1, 1, 1},
        {219, 2, 1, 1, 227, 2, 1, 1},
        {219, 1, 2, 1, 227, 1, 1, 2},
        {219, 2, 2, 1, 227, 2, 1, 2},
        {219, 1, 3, 1, 226, 1, 1, 1},
        {219, 2, 3, 1, 226, 2, 1, 1},
        {221, 1, 1, 1, 229, 1, 1, 1},
        {222, 1, 1, 1, 230, 1, 1, 1},
        {222, 1, 2, 1, 230, 1, 1, 2},
        {222, 2, 1, 1, 229, 1, 1, 2},
        {222, 2, 2, 1, 229, 1, 1, 3},
        {224, 1, 1, 1, 233, 1, 1, 1},
        {224, 1, 2, 1, 233, 1, 1, 2},
        {224, 2, 1, 1, 233, 1, 1, 3},
        {224, 2, 2, 1, 233, 1, 1, 4},
        {224, 3, 1, 1, 233, 1, 1, 5},
        {224, 3, 2, 1, 233, 1, 1, 6},
        {225, 1, 1, 1, 234, 1, 1, 1},
        {225, 1, 2, 1, 234, 2, 1, 1},
        {227, 1, 1, 1, 240, 1, 1, 1},
        {227, 1, 2, 1, 240, 1, 2, 1},
        {227, 2, 1, 1, 240, 1, 1, 2},
        {227, 2, 2, 1, 240, 1, 2, 2},
        {227, 3, 1, 1, 240, 1, 1, 3},
        {227, 3, 2, 1, 240, 1, 2, 3},
        {227, 4, 1, 1, 239, 1, 1, 1},
        {227, 4, 2, 1, 239, 1, 2, 1},
        {228, 1, 1, 1, 241, 1, 1, 1},
        {228, 1, 2, 1, 241, 1, 2, 1},
        {228, 1, 3, 1, 241, 1, 2, 1},
        {228, 2, 1, 1, 241, 1, 1, 2},
        {228, 2, 2, 1, 241, 1, 2, 2},
        {228, 2, 3, 1, 241, 1, 2, 2},
        {228, 3, 1, 1, 241, 1, 1, 3},
        {228, 3, 2, 1, 241, 1, 2, 3},
        {228, 3, 3, 1, 241, 1, 2, 3},
        {229, 1, 1, 1, 242, 1, 1, 1},
        {229, 1, 2, 1, 242, 1, 2, 1},
        {230, 1, 1, 1, 243, 1, 1, 1},
        {230, 1, 2, 1, 243, 1, 2, 1},
        {230, 1, 3, 1, 243, 1, 2, 1},
        {230, 2, 1, 1, 243, 1, 1, 2},
        {230, 2, 2, 1, 243, 1, 2, 2},
        {230, 2, 3, 1, 243, 1, 2, 2},
        {231, 1, 1, 1, 244, 1, 1, 1},
        {231, 1, 2, 1, 244, 1, 2, 1},
        {231, 1, 3, 1, 244, 1, 2, 1},
        {231, 2, 1, 1, 244, 1, 1, 2},
        {231, 2, 2, 1, 244, 1, 2, 2},
        {231, 2, 3, 1, 244, 1, 2, 2},
        {231, 3, 1, 1, 244, 1, 1, 3},
        {231, 3, 2, 1, 244, 1, 2, 3},
        {231, 3, 3, 1, 244, 1, 2, 3},
        {231, 4, 1, 1, 244, 1, 1, 4},
        {231, 4, 2, 1, 244, 1, 2, 4},
        {231, 4, 3, 1, 244, 1, 2, 4},
        {232, 1, 1, 1, 245, 1, 1, 1},
        {232, 1, 2, 1, 245, 1, 2, 1},
        {232, 1, 3, 1, 245, 1, 2, 1},
        {232, 2, 1, 1, 245, 1, 1, 2},
        {232, 2, 2, 1, 245, 1, 2, 2},
        {232, 2, 3, 1, 245, 1, 2, 2},
        {232, 3, 1, 1, 245, 1, 1, 3},
        {232, 3, 2, 1, 245, 1, 2, 3},
        {232, 3, 3, 1, 245, 1, 2, 3},
        {232, 4, 1, 1, 245, 1, 1, 4},
        {232, 4, 2, 1, 245, 1, 2, 4},
        {232, 4, 3, 1, 245, 1, 2, 4},
        {239, 1, 1, 1, 246, 1, 1, 1},
        {239, 1, 2, 1, 246, 1, 2, 1},
        {239, 1, 3, 1, 246, 1, 2, 1},
        {239, 2, 1, 1, 246, 1, 1, 2},
        {239, 2, 2, 1, 246, 1, 2, 2},
        {239, 2, 3, 1, 246, 1, 2, 2},
        {240, 1, 1, 1, 247, 1, 1, 1},
        {240, 1, 2, 1, 247, 1, 2, 1},
        {240, 1, 3, 1, 247, 1, 2, 1},
        {240, 2, 1, 1, 247, 1, 1, 2},
        {240, 2, 2, 1, 247, 1, 2, 2},
        {240, 2, 3, 1, 247, 1, 2, 2},
        {240, 3, 1, 1, 247, 1, 1, 3},
        {240, 3, 2, 1, 247, 1, 2, 3},
        {240, 3, 3, 1, 247, 1, 2, 3},
        {243, 1, 1, 1, 251, 1, 1, 1},
        {243, 2, 1, 1, 251, 1, 1, 2},
        {249, 1, 1, 1, 252, 1, 2, 1},
        {249, 1, 2, 1, 252, 1, 2, 3},
        {249, 2, 1, 1, 253, 1, 2, 1},
        {249, 2, 2, 1, 253, 1, 2, 3},
        {249, 3, 1, 1, 254, 1, 2, 1},
        {249, 3, 2, 1, 255, 1, 2, 1},
        {249, 4, 1, 1, 256, 1, 2, 1},
        {249, 4, 2, 1, 256, 1, 2, 5},
        {250, 1, 1, 1, 252, 1, 2, 2},
        {250, 2, 1, 1, 253, 1, 2, 2},
        {250, 3, 1, 1, 257, 1, 2, 1},
        {250, 4, 1, 1, 257, 1, 2, 2},
        {251, 1, 1, 1, 258, 1, 1, 1}
    };
    
    // An array that is used to fill in the phase3Map hashtable
    // Each line consists of 
    // { old_code, old_fill, old_var, new_code, new_fill,
    //      new_var}
    // NOTE:  the values of fill, and var (variation or ahead)
    //      in this table start counting at one, while the actual values start
    //      counting at zero.  Subtract one from all fill, and var values
    //      when constructing the hashtable.  These values are the ones Valerie
    //      sees when using her symbol table editor.
    // NOTE:  unlike phase2Map, these transforms pertain to all rotations (rotations
    //      are unchanged)
    private static final int phase3MapInit[][] = {
        {116, 2, 1, 116, 1, 2},
        {116, 3, 1, 116, 1, 3},
        {116, 4, 1, 116, 1, 4},
        {117, 2, 1, 117, 1, 2},
        {117, 3, 1, 117, 1, 3},
        {117, 4, 1, 117, 1, 4},
        {203, 2, 1, 203, 1, 2},
        {203, 3, 1, 203, 1, 3},
        {203, 4, 1, 203, 1, 4},
        {210, 1, 1, 211, 1, 1},
        {210, 2, 1, 211, 1, 2},
        {220, 1, 1, 228, 1, 1},
        {220, 2, 1, 228, 1, 2},
        {220, 3, 1, 228, 1, 3},
        {220, 4, 1, 228, 1, 4},
        {223, 1, 1, 232, 1, 1},
        {226, 1, 1, 235, 1, 1},
        {226, 2, 1, 235, 1, 2},
        {233, 1, 1, 236, 1, 1},
        {233, 2, 1, 236, 1, 2},
        {233, 3, 1, 236, 1, 4},
        {234, 1, 1, 236, 1, 3},
        {235, 1, 1, 236, 1, 5},
        {236, 1, 1, 237, 1, 1},
        {236, 2, 1, 237, 1, 2},
        {236, 3, 1, 237, 1, 4},
        {237, 1, 1, 237, 1, 3},
        {238, 1, 1, 237, 1, 5},
        {241, 1, 1, 248, 1, 1},
        {241, 2, 1, 248, 1, 2},
        {241, 3, 1, 248, 1, 3},
        {242, 1, 1, 249, 1, 1},
        {242, 2, 1, 249, 1, 2},
        {242, 3, 1, 249, 1, 3},
        {244, 1, 1, 250, 1, 1},
        {244, 2, 1, 250, 1, 2},
        {244, 3, 1, 250, 1, 3},
        {245, 1, 1, 259, 1, 1},
        {245, 2, 1, 259, 1, 2},
        {245, 3, 1, 259, 1, 3},
        {245, 4, 1, 259, 1, 4},
        {245, 5, 1, 259, 1, 5},
        {245, 6, 1, 259, 1, 6},
        {246, 1, 1, 260, 1, 1},
        {246, 2, 1, 260, 1, 2},
        {246, 3, 1, 260, 1, 3},
        {246, 4, 1, 260, 1, 4},
        {247, 1, 1, 261, 1, 1},
        {247, 2, 1, 261, 1, 2},
        {247, 3, 1, 261, 1, 3},
        {247, 4, 1, 261, 1, 4},
        {248, 1, 1, 262, 1, 1},
        {248, 2, 1, 262, 1, 2},
        {248, 3, 1, 262, 1, 3},
        {248, 4, 1, 262, 1, 4}
    };
    
    // change values in start and end to conform to new symbol order.
    // These values indicate symbol code ranges that need special treatment
    // when being translated from the old symbol code order to the new (phase 1)
    // order.
    // 12-17-98 LSP
    private static final int[] start = { 118, 124, 133, 141, 148, 163, 167, 172, 176};
    private static final int[] end   = { 118, 129, 135, 144, 151, 165, 169, 174, 178};
    
    static {
        hash1 = new Hashtable();
        int i;
        for (int j = 0; j < start.length; j++) {
            for (i = start[j]; i <= end[j]; i++) {
                hash1.put(new Integer(i), new Integer(i));
            }
        }
        
        // add phase2Map to complete mapping of old symcodes(fills, rots, vars) to
        // new symcodes etc.
        phase2Map = new Hashtable();
        int key, value;
        for (int j = 0; j < phase2MapInit.length; j++) {
            key = 1000 * phase2MapInit[j][0] +       // old sym code
                   100 * (phase2MapInit[j][1] - 1) + // old fill
                    10 * (phase2MapInit[j][2] - 1) + // old rotation
                         (phase2MapInit[j][3] - 1);  // old variation or ahead
            value = 1000 * phase2MapInit[j][4] +       // new sym code
                     100 * (phase2MapInit[j][5] - 1) +  // new fill
                      10 * (phase2MapInit[j][6] - 1) +  // new rotation
                           (phase2MapInit[j][7] - 1);   // new variation
            phase2Map.put(new Integer(key), new Integer(value));
        }

        // add phase3Map to complete mapping of old symcodes(fills, vars) to
        // new symcodes etc.  All rotations are unaffected in phase 3
        phase3Map = new Hashtable();
        for (int j = 0; j < phase3MapInit.length; j++) {
            key = 100 * phase3MapInit[j][0] +       // old sym code
                   10 * (phase3MapInit[j][1] - 1) + // old fill
                        (phase3MapInit[j][2] - 1);  // old variation or ahead
            value = 100 * phase3MapInit[j][3] +       // new sym code
                     10 * (phase3MapInit[j][4] - 1) +  // new fill
                          (phase3MapInit[j][5] - 1);   // new variation
            phase3Map.put(new Integer(key), new Integer(value));
        }
    }


    private int code;
    private boolean flop;
    private byte rot;
    private byte ahead;
    private byte fill;
    private int centerx = 0;
    private int centery = 0;

//    public SignSymbol(int code, int opt, byte y, byte x) {
//        this(code, opt, y, x, false);
//    }
   
    public SignSymbol(int code, int opt, byte y, byte x, boolean version4_0) {
        int _rot      = (opt & 0x7);
        int _ahead    = Math.max(0,((opt >> 3) & 0x3) - 1);
        int _fill     = ((opt >> 5) & 0x3) + ((((opt >> 3) & 0x3) == 0)? 4 : 0);
        boolean _flop = ((opt>>7) == 1);  
         
        // change values of code to conform to new symbol order
        // 12-17-98 LSP
        if (version4_0) {
            code = mapOldToNew[code];
            if (hash1.contains(new Integer(code))) {
                int temp = _fill;
                _fill = _ahead;
                _ahead = temp;
                boolean changedArrow = false;
                if (_ahead == 1) {
                    _ahead = 0;
                    changedArrow = true;
                }
                if (!changedArrow && _ahead == 0) {
                    _ahead = 1;
                }
            }
            else if (code == 191 || code == 36) {
                _fill -= 1;
                if (_fill < 0) _fill += 4;
            }
            
            if (_flop && _rot > 0 && code != 154) {
                _rot = Symbol.getSymbol(code).getNumRotations(_ahead, _fill) - _rot;
            }
            
            int key2 = 1000 * code + 100 * _fill + 10 * _rot + _ahead;
            Integer Ikey2 = new Integer(key2);
            
            int key3 = 100 * code + 10 * _fill + _ahead;
            Integer Ikey3 = new Integer(key3);

            if (phase2Map.containsKey(Ikey2)) {
                int value = ((Integer) phase2Map.get(Ikey2)).intValue();
                code = value / 1000;
                _fill = (value % 1000) / 100;
                _rot = (value % 100) / 10;
                _ahead = value % 10;
            } else if (phase3Map.containsKey(Ikey3)) {
                int value = ((Integer) phase3Map.get(Ikey3)).intValue();
                code = value / 100;
                _fill = (value % 100) / 10;
                _ahead = value % 10;
            }


        }  // if (version4_0)
        init(code, _rot, _ahead, _fill, _flop, (x*2), (y*2));
    }

    public SignSymbol(int code, int rot, int ahead, int fill, boolean flop) {
        this(code, rot, ahead, fill, flop, 0, 0);
    }

    public SignSymbol(int code, int rot, int ahead, int fill, boolean flop,
                      int x, int y) {
        init(code, rot, ahead, fill, flop, x, y);
    }
    
    private void init(int code, int rot, int ahead, int fill, boolean flop,
                      int x, int y) {
        this.code = code;
        this.rot = (byte)rot;
        this.flop = flop;
        this.ahead = (byte)ahead;
        this.fill = (byte)fill;
        this.x = x;
        this.y = y;
        ImageCoords coords = null;
        try {
            coords = Symbol.getSymbol(code).getCoords(ahead,
                                              this.fill, this.rot,
                                              this.flop, false);
        }
        catch (Exception e) {
            System.err.println("\n\nException in SignSymbol.init, with\n" +
                "symcode: " + code + ", fill: " + fill + ", rotation: " + rot + 
                ", variation (ahead): " + ahead + ", flop: " + flop + "\n");
        }
        if (coords != null) {
            centerx = coords.centerPtH;
            centery = coords.centerPtV;
        }

        resetSize();
    }

    public SignSymbol(SignSymbol rhs) {
        super(rhs);
        code = rhs.code;
        flop = rhs.flop;
        rot = rhs.rot;
        ahead = rhs.ahead;
        fill = rhs.fill;
        centerx = rhs.centerx;
        centery = rhs.centery;
    }

    public Object clone() {
        SignSymbol newSignSymbol = new SignSymbol(this);
        return newSignSymbol;
    }

    public int hashCode() {
        return code + (rot << 8) + (ahead << 11) + (fill << 13) + ((flop?1:0) << 15);
    }

    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        SignSymbol rhs = (SignSymbol)obj;
        if (rhs.code != code ||
            rhs.flop != flop ||
            rhs.rot != rot ||
            rhs.ahead != ahead ||
            rhs.fill != fill) {
            return false;
        }
        return true;
    }

    public int getNumVariations() {
        return Symbol.getSymbol(code).getNumVariations();
    }

    public int getNumFills() {
        return Symbol.getSymbol(code).getNumFills(ahead);
    }

    public int getNumRotations() {
        return Symbol.getSymbol(code).getNumRotations(ahead, fill);
    }

    public int getNumFlops() {
        return Symbol.getSymbol(code).getNumFlops(ahead, fill);
    }

    public int getCode() {
        return code;
    } 
    
    public int getRot() {
        return rot;
    } 
    
    public void setRot(int i) {
        rot = (byte)i;
    }
    
    public void setRotation(int i) {
        rot = (byte)i;
        //WHY?
        resetSize();
    }
    
    public void setFill(int i) {
        fill = (byte)i;
        //resetSize();
    }

    public void resetSize() {
        if (ahead >= getNumVariations()) {
            ahead = 0;
        }
        if (fill >= getNumFills()) {
            fill = 0;
        }
        if (rot >= getNumRotations()) {
            rot = 0;
        } else if (rot < 0) {
            rot = (byte)(getNumRotations()-1);
        }
        if (getNumFlops() == 0) {
            //flop = false; //NO!NO!NO!
        }
        ImageCoords coords = Symbol.getSymbol(code).getCoords(
                                              ahead, this.fill, this.rot,
                                              this.flop, false);    

        x += centerx - coords.centerPtH;
        y += centery - coords.centerPtV;
        
        centerx = coords.centerPtH;
        centery = coords.centerPtV;
        super.resetSize();
    }
    
    protected int calcWidth() {
        ImageCoords coords = Symbol.getSymbol(code).getCoords(
                                              ahead, fill, rot,
                                              flop, false);
        return coords.framePtH;
    }
    
    protected int calcHeight() {
        ImageCoords coords = Symbol.getSymbol(code).getCoords(
                                              ahead, fill, rot,
                                              flop, false);
        return coords.framePtV;
    }

    public void incrementRotation(boolean clockwise) {
        if ( (clockwise && flop) || !(clockwise || flop) ) {
            rot++;
        } else {
            rot--;
            if (rot < 0) {
                rot += getNumRotations();
            }
        }
        resetSize();
    }

    public void incrementFill(boolean clockwise) {
        if (clockwise) {
            fill--;
            if (fill < 0) {
                fill += getNumFills();
            }
        }
        else {
            fill++;
        }
        resetSize();
    }

    public void changeFlop() {
        flop = !flop;
        resetSize();
    }

    public void incrementVariation(boolean clockwise) {
        if (clockwise) {
            ahead--;
            if (ahead < 0) {
                ahead += getNumVariations();
            }
        }
        else {
            ahead++;
        }
        resetSize();
    }

    public void paint(GadgetGraphics g, int x, int y, int offsetx, int offsety,
                      Bitmap bitmap, int[] pixels, PaintInfo paintInfo) {
        int xoffset = this.x;
        int yoffset = this.y;
        boolean thumbNail = false;
        if (g.getFont() == SignFont.small) {
            thumbNail = true;
            xoffset = ((xoffset) / 2);
            yoffset = ((yoffset) / 2);
        }
        Symbol.getSymbol(code).paint(g,
                     x+xoffset,y+yoffset,ahead,fill,rot,flop,thumbNail,
                     bitmap,xoffset+offsetx,yoffset+offsety, pixels, paintInfo );
    }

    public String toString() {
        return "code="+code+" y="+y+" x="+x+
               " flop="+flop+" rot="+rot+" arrowhead="+ahead+" fill="+fill;
    }


}

